package edu.fichte.pbi.vm.var;

import java.util.ArrayList;

import edu.fichte.pbi.vm.var.format.Formatter;

/**
 * A variable which consists of individually accessible bits.
 * 
 * @author Elias Groll
 * @version 2.15
 */
public class ConglomerateVariable implements IVariable {
	
	/**
	 * Listener which fires when Variable is/gets changed
	 */
	protected IVariableListener iVariableListener;
	
	/**
	 * if constructed out of other ConglomerateVariables,
	 * contains its parents (you can use this to fire
	 * state changes to the parents when the child is changed)
	 */
	private ConglomerateVariable[] childs;
	
	protected ArrayList<ConglomerateVariable> parents;
	
  /**
   * @param bits
   *          the bit[] which should be used to set a value.
   */
  public ConglomerateVariable(Bit[] bits) {
    this.bits = bits;
    for (int i = 0; i < bits.length; i++) {
      if (this.bits[i] == null) {
        this.bits[i] = new Bit(0);
      }
    }
    this.maximum = (long) 1 << this.bits.length;
  }
  
  /**
   * Allows the Construction of a ConglomerateVariable out
   * of ConglomerateVariables so it can properly fire state
   * changes and push them to the GUI
   * @param conglomerates to use when constructing the ConglomerateVariable
   */
  public ConglomerateVariable(ConglomerateVariable[] conglomerates) {
	  childs = conglomerates;
	  int totalBits = 0;
	  
	  for (int i = 0; i < conglomerates.length; i++) {
		  totalBits += conglomerates[i].getBits().length;
		  if (conglomerates[i].parents != null) {
			  conglomerates[i].parents.add(this);
		  }
	  }
	  
	  this.bits = new Bit[totalBits];
	  for (int i = 0; i < conglomerates.length; i++) {
		  for (int j = 0; j < conglomerates[i].getBits().length; j++) {
			  int index = (i*conglomerates[i].getBits().length) + j;
			  bits[index] = conglomerates[i].getBit(j);
		  }
	  }
  }

  /**
   * The bits which the Variable uses to store its value
   */
  protected Bit[] bits;
  protected long maximum;

  private String cutToByteLength(String byteAsString) {
    String retVal = "";
    if (byteAsString.length() > bits.length) {
      int index = byteAsString.length() - bits.length;
      retVal = byteAsString.substring(index);
    } else if (byteAsString.length() < bits.length) {
      int count = bits.length - byteAsString.length();
      retVal = byteAsString;
      for (int i = 0; i < count; i++) {
        retVal = "0" + retVal;
      }
    } else {
      retVal = byteAsString;
    }

    return retVal;
  }

  
  public String toString() {
    String value = Formatter.convertToHexString(getValue());
    return value;
  }

  /**
   * Get an individually accessible bit from the variable.
   * 
   * @param num
   *          the index of the bit in the variable
   * @return the bit at the specified index
   */
  public Bit getBit(int num) {
    if (num >= 0 && num < bits.length) {
      return bits[num];
    } else {
      return null;
    }
  }

  
  public long getValue() {
    long retVal = 0;
    long comp = 1;
    for (int i = bits.length - 1; i >= 0; i--) {
      if (bits[i].getValue() == 1) {
        retVal += comp;
      }
      comp *= 2;
    }
    return retVal;

  }

  
  public void setValue(long value) {
//	  System.out.println("Value to be set: " + value + " / max bits size: " + bits.length);
    long val = value;
    while (val < 0) {
      val = (maximum - (value * -1));
    }
    String byteAsString = "";
    byteAsString = Long.toBinaryString(val);
//    System.out.println("Raw Byte Binary String " + byteAsString);
    byteAsString = cutToByteLength(byteAsString);
//    System.out.println("Byte String cut to length " + byteAsString);
    for (int index = 0; index < bits.length; index++) {
//    	System.out.println("Set bit" + index + " to " + byteAsString.charAt(index));
      bits[index].setValue(Integer.parseInt(String.valueOf(byteAsString.charAt(index)))); //TODO Fails
    }
    fireStateChange();
    if (childs != null) {
    	for (ConglomerateVariable child : childs) {
    		child.fireStateChange();
    	}
    }
    if (parents != null) {
    	for (ConglomerateVariable parent : parents) {
    		parent.fireStateChange();
    	}
    }
  }

  public Bit[] getBits() {
    return bits;
  }

  public long getMaximum() {
    return maximum;
  }
  
  public void fireStateChange() {
	  if (iVariableListener != null) {
		  iVariableListener.variableChanged(this);
	  }
  }

  
	@Override
	public void setVariableListener(IVariableListener iVariableListener) {
		this.iVariableListener = iVariableListener;
	}
}
